# sndid/jackmain.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
from sndidjack.sndidjack import SndIdJack
from sndidjack.config import Config


def main():
    config = Config.load_config()
    sndidjack = SndIdJack(config)
    sndidjack.main()


if __name__ == "__main__":
    main()



# sndid/cli.py
# Copyright 2023, 2024 Jeff Moe <moe@spacecruft.org>
# Copyright 2022, 2023, Joe Weiss <joe.weiss@gmail.com>
# Licensed under the Apache License, Version 2.0
import argparse
import io
from birdnetlib import Recording
from birdnetlib.analyzer import Analyzer
from contextlib import redirect_stdout, redirect_stderr
from datetime import datetime, date


def parse_args():
    TODAY = date.today()
    YEAR = TODAY.year
    MONTH = TODAY.month
    DAY = TODAY.day

    parser = argparse.ArgumentParser(
        description="Run sndid to identify bird species in audio recordings."
    )
    parser.add_argument(
        "-i",
        "--input",
        help="Input filename to process (default samples/sample.wav)",
        type=str,
        required=False,
        default="samples/sample.wav",
    )
    parser.add_argument(
        "-t",
        "--latitude",
        help="Latitude (default 40.57)",
        type=str,
        required=False,
        default="40.57",
    )
    parser.add_argument(
        "-n",
        "--longitude",
        help="Longitude (default -105.23)",
        type=str,
        required=False,
        default="-105.23",
    )
    parser.add_argument(
        "-y",
        "--year",
        help="Year (default today)",
        type=int,
        required=False,
        default=YEAR,
    )
    parser.add_argument(
        "-m",
        "--month",
        help="Month (default today)",
        type=int,
        required=False,
        default=MONTH,
    )
    parser.add_argument(
        "-d",
        "--day",
        help="Day (default today)",
        type=int,
        required=False,
        default=DAY,
    )
    parser.add_argument(
        "-c",
        "--confidence",
        help="Minimum Confidence (default 0.50)",
        type=float,
        required=False,
        default="0.50",
    )

    parser.add_argument(
        "-l",
        "--list",
        help="Output as human readable list not terse (default True)",
        action=argparse.BooleanOptionalAction,
        type=bool,
        required=False,
        default=True,
    )

    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    today = datetime.today()
    args.year = args.year or today.year
    args.month = args.month or today.month
    args.day = args.day or today.day

    output_buffer = io.StringIO()
    with redirect_stdout(output_buffer), redirect_stderr(output_buffer):
        custom_labels_path = "models/analyzer/BirdNET_GLOBAL_6K_V2.4_Labels.txt"
        # Model used 2023-Q4 - 2024-Q1.
        custom_model_path = "models/analyzer/BirdNET_GLOBAL_6K_V2.4_Model_FP32.tflite"
        # Newer than above, but not default in birdnetlib because it hangs. XXX
        # custom_model_path = "models/analyzer/BirdNET_GLOBAL_6K_V2.4_MData_Model_V2_FP16.tflite"
        analyzer = Analyzer(
            classifier_labels_path=custom_labels_path,
            classifier_model_path=custom_model_path,
        )

        recording = Recording(
            analyzer,
            args.input,
            lat=args.latitude,
            lon=args.longitude,
            date=datetime(year=args.year, month=args.month, day=args.day),
            min_conf=args.confidence,
        )

        recording.analyze()

    if args.list:
        i = 0
        species_sort = ""
        for i in range(0, len(recording.detections)):
            species_sort = species_sort + (
                recording.detections[i]["common_name"]
                + ", "
                + recording.detections[i]["scientific_name"]
                + ", "
                + str(recording.detections[i]["start_time"])
                + ", "
                + str(recording.detections[i]["end_time"])
                + ", "
                + str(recording.detections[i]["confidence"])
                + "\n"
            )
        species_out = sorted(species_sort.split("\n"))

        i = 0
        for i in range(1, len(species_out)):
            print(species_out[i])

    else:
        print(recording.detections)


if __name__ == "__main__":
    main()



# sndid/audiodetectionwriter.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import queue
import threading
from .detectionswriter import DetectionsWriter


class AudioDetectionWriter:
    def __init__(self, detections_file, detections_dir):
        logging.debug("Initialize AudioDetectionWriter")
        self.detections_writer = DetectionsWriter(detections_file, detections_dir)
        self.detection_queue = queue.Queue()
        self._stop_event = threading.Event()
        self._writer_thread = threading.Thread(target=self._start_writer_thread)
        self._writer_thread.daemon = True
        self._writer_thread.start()

    def write_detections(
        self, recording_data, detections, detection_timestamp, input_index
    ):
        logging.debug("Putting data onto the queue")
        self.detection_queue.put(
            (recording_data, detections, detection_timestamp, input_index)
        )
        logging.debug("write_detections")

    def _write_detections_loop(self):
        while not self._stop_event.is_set():
            try:
                recording_data, detections, detection_timestamp, input_index = (
                    self.detection_queue.get(timeout=1)
                )
                logging.debug("Got data from the queue")
            except queue.Empty:
                continue
            self.detections_writer.write_detections(
                detections, detection_timestamp, input_index
            )
            self.detections_writer.write_audio(
                recording_data, detections, detection_timestamp, input_index
            )
            logging.debug("Wrote data to file")
            self.detection_queue.task_done()
        self.detection_queue.join()

    def _start_writer_thread(self):
        logging.basicConfig(level=logging.DEBUG)
        self._write_detections_loop()

    def stop(self):
        self._stop_event.set()
        self._writer_thread.join()
        logging.debug("Writer stopped")



# sndid/list.py
# Copyright 2023, 2024 Jeff Moe <moe@spacecruft.org>
# Copyright 2022, 2023, Joe Weiss <joe.weiss@gmail.com>
# Licensed under the Apache License, Version 2.0

import argparse
import os
from birdnetlib.species import SpeciesList
from contextlib import redirect_stdout, redirect_stderr
from datetime import date

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


def parse_arguments():
    today = date.today()
    parser = argparse.ArgumentParser(
        description="sndid-list -- List available species. By default return all. Limit list with date and location options."
    )
    parser.add_argument(
        "-a", "--latitude", help="Latitude (default 40.57)", type=str, default="40.57"
    )
    parser.add_argument(
        "-o",
        "--longitude",
        help="Longitude (default -105.23)",
        type=str,
        default="-105.23",
    )
    parser.add_argument(
        "-y", "--year", help="Year (default today)", type=int, default=today.year
    )
    parser.add_argument(
        "-m", "--month", help="Month (default today)", type=int, default=today.month
    )
    parser.add_argument(
        "-d", "--day", help="Day (default today)", type=int, default=today.day
    )
    parser.add_argument(
        "-t",
        "--threshold",
        help="Threshold value for species list (default 0.0)",
        type=float,
        default=0.0,
    )
    return parser.parse_args()


def get_species_list(args):
    species_local = SpeciesList()

    date_obj = (
        date(year=args.year, month=args.month, day=args.day)
        if args.year and args.month and args.day
        else None
    )
    lat = float(args.latitude) if args.latitude else None
    lon = float(args.longitude) if args.longitude else None

    filtered_species_list = species_local.return_list(
        lat=lat, lon=lon, date=date_obj, threshold=args.threshold
    )

    return filtered_species_list


def format_species_list(species_list):
    return sorted(
        f"{species['common_name']}, {species['scientific_name']}"
        for species in species_list
    )


def main():
    args = parse_arguments()

    with open(os.devnull, "w") as devnull:
        with redirect_stdout(devnull), redirect_stderr(devnull):
            species_list = get_species_list(args)

    species_out = format_species_list(species_list)

    for species in species_out:
        print(species)


if __name__ == "__main__":
    main()



# sndid/textdetectionwriter.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import queue
import threading
from .textwriter import TextWriter


class TextDetectionWriter:
    def __init__(self, detections_file, detections_dir):
        logging.debug("Initializing TextDetectionWriter")
        self.text_writer = TextWriter(detections_file)
        self.detection_queue = queue.Queue()
        self._stop_event = threading.Event()
        self._writer_thread = threading.Thread(target=self._write_detections_loop)
        self._writer_thread.daemon = True
        self._writer_thread.start()
        logging.debug("TextDetectionWriter thread started")

    def write_detections(self, detections, detection_timestamp, input_index):
        self.detection_queue.put((detections, detection_timestamp, input_index))
        logging.debug(
            f"TextDetectionWriter detections written to queue: {len(detections)}"
        )

    def _write_detections_loop(self):
        while not self._stop_event.is_set():
            try:
                detections, detection_timestamp, input_index = self.detection_queue.get(
                    timeout=1
                )
                logging.debug("TextDetectionWriter received detections from queue")
            except queue.Empty:
                continue
            self.text_writer.write_detections(
                detections, detection_timestamp, input_index
            )
            logging.debug(
                f"TextDetectionWriter detections written to file: {len(detections)}"
            )

    def stop(self):
        logging.debug("Stopping TextDetectionWriter")
        self._stop_event.set()
        self._writer_thread.join()



# sndid/alsa.py
# Copyright 2023, 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0

import argparse
import os


def main():
    parser = argparse.ArgumentParser(description="Run sndid-alsa")
    parser.add_argument(
        "-i",
        "--ip",
        help="Server IP address (default 127.0.0.1)",
        type=str,
        required=False,
        default="127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Server network port (default 9988)",
        type=str,
        required=False,
        default="9988",
    )
    parser.add_argument(
        "-r",
        "--rate",
        help="Rate in Hertz (default 48000)",
        type=str,
        required=False,
        default=48000,
    )
    parser.add_argument(
        "-t",
        "--time",
        help="Length of segments in seconds (default 10)",
        type=str,
        required=False,
        default=10,
    )

    args = parser.parse_args()
    IP = args.ip
    PORT = args.port
    RATE = str(args.rate)
    TIME = str(args.time)

    CMD = (
        "arecord --rate "
        + RATE
        + " -f FLOAT_LE --max-file-time "
        + TIME
        + " | nc "
        + IP
        + " "
        + PORT
    )
    print("Streaming ALSA in to " + IP + ":" + PORT)
    os.system(CMD)


if __name__ == "__main__":
    main()



# sndid/textwriter.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import os


class TextWriter:
    def __init__(self, detections_file):
        self.detections_file = detections_file
        logging.debug("TextWriter object created.")

    def write_detections(self, detections, detection_timestamp, input_index):
        detections_dir = os.path.dirname(self.detections_file)
        logging.debug(f"Creating directory: {detections_dir}")
        os.makedirs(detections_dir, exist_ok=True)

        with open(self.detections_file, "a") as file:
            for d in detections:
                file.write(
                    f"{detection_timestamp}, {d['common_name']}, {d['scientific_name']}, {d['confidence']}, Input {input_index + 1}\n"
                )
        logging.debug(f"Detections written to text file: {self.detections_file}")



# sndid/client.py
# Copyright 2023, 2024 Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import argparse
import os


def main():
    parser = argparse.ArgumentParser(description="Run sndid-client")
    parser.add_argument(
        "-i",
        "--ip",
        help="Server IP address (default 127.0.0.1)",
        type=str,
        required=False,
        default="127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Server network port (default 9988)",
        type=int,
        required=False,
        default="9988",
    )
    parser.add_argument(
        "-f",
        "--file",
        help="Input filename to process (default samples/mono.wav)",
        type=str,
        required=False,
        default="samples/mono.wav",
    )

    args = parser.parse_args()
    IP = args.ip
    PORT = str(args.port)
    FILE = args.file

    CMD = "cat" + " " + FILE + " | nc -q 0 " + IP + " " + PORT
    print("Sending " + FILE + " to " + IP + ":" + PORT)
    os.system(CMD)


if __name__ == "__main__":
    main()



# sndid/__init__.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import importlib.metadata

try:
    __version__ = importlib.metadata.version("sndid")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.1"



# sndid/jackcli.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import argparse
import logging
import sys
from sndid.sndidjack import SndIdJack
from sndid.config import Config
from .__version__ import __version__


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="sndid-jack - AI processing JACK Audio Connection Kit."
    )
    parser.add_argument("--config_file", "-c", help="Configuration File name")
    parser.add_argument(
        "--log-level",
        "-L",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Logging level.",
        type=str.upper,
    )
    parser.add_argument("--n-inputs", "-n", type=int, help="Number of inputs")
    parser.add_argument("--segment-length", "-l", type=int, help="Segment length")
    parser.add_argument("--min-confidence", "-m", type=float, help="Minimum confidence")
    parser.add_argument("--lat", "-y", type=float, help="Latitude")
    parser.add_argument("--lon", "-x", type=float, help="Longitude")
    parser.add_argument("--output-dir", "-o", help="Output directory for detections")
    parser.add_argument(
        "--version", "-v", action="version", version=f"%(prog)s {__version__}"
    )

    return parser.parse_args()


def main():
    args = parse_arguments()
    config_file = args.config_file
    config = Config.load_config(config_file)

    if not config_file:
        config = Config.set_defaults({})

    for key, value in vars(args).items():
        if value is not None:
            config[key] = value

    logging.basicConfig(
        level=config["log_level"],
        format="%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stdout,
    )

    logging.debug("Starting SndIdJack with configuration: %s", config)

    sndidjack = SndIdJack(config)
    sndidjack.main()


if __name__ == "__main__":
    main()



# sndid/server.py
# Copyright 2023, 2024 Jeff Moe <moe@spacecruft.org>
# Copyright 2022, 2023, Joe Weiss <joe.weiss@gmail.com>
# Licensed under the Apache License, Version 2.0
import argparse
import birdnetlib.wavutils as wavutils
import datetime
import functools
import socketserver
import io
import logging
from birdnetlib import RecordingBuffer
from birdnetlib.analyzer import Analyzer
from contextlib import redirect_stdout, redirect_stderr


class TCPHandler(socketserver.StreamRequestHandler):
    def __init__(self, *args, **kwargs):
        self.lat = kwargs.get("lat", None)
        self.lon = kwargs.get("lon", None)
        self.year = kwargs.get("year", None)
        self.month = kwargs.get("month", None)
        self.day = kwargs.get("day", None)
        self.confidence = kwargs.get("confidence", None)
        super().__init__(*args)

    def handle(self):
        f = io.StringIO()
        with redirect_stdout(f), redirect_stderr(f):
            analyzer = Analyzer()
        # Read WAV data from the socket
        for rate, data in wavutils.bufferwavs(self.rfile):
            with redirect_stdout(f), redirect_stderr(f):
                # Make a RecordingBuffer with buffer and rate
                recording = RecordingBuffer(
                    analyzer,
                    data,
                    rate,
                    lat=self.lat,
                    lon=self.lon,
                    date=datetime.datetime(
                        year=self.year, month=self.month, day=self.day
                    ),
                    min_conf=self.confidence,
                )
            with redirect_stdout(f), redirect_stderr(f):
                recording.analyze()
            i = 0
            detections_sort = ""
            for i in range(0, len(recording.detections)):
                detections_sort = detections_sort + (
                    recording.detections[i]["common_name"]
                    + ", "
                    + recording.detections[i]["scientific_name"]
                    + ", "
                    + str(recording.detections[i]["confidence"])
                    + "\n"
                )
            detections_out = sorted(detections_sort.split("\n"))
            i = 0
            for i in range(1, len(detections_out)):
                n = datetime.datetime.now(datetime.timezone.utc).astimezone()
                print(n, detections_out[i])
                logging.info(str(n) + " " + str(detections_out[i]))


def initialize_logging():
    logging.basicConfig(
        filename="sndid.log",
        encoding="utf-8",
        format="%(message)s",
        level=logging.DEBUG,
    )


def start_server():
    TODAY = datetime.date.today()
    YEAR = TODAY.year
    MONTH = TODAY.month
    DAY = TODAY.day

    parser = argparse.ArgumentParser(description="Run sndid-server")
    parser.add_argument(
        "-i",
        "--ip",
        help="Server IP address (default 127.0.0.1)",
        type=str,
        required=False,
        default="127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Server network port (default 9988)",
        type=int,
        required=False,
        default="9988",
    )
    parser.add_argument(
        "-t",
        "--latitude",
        help="Latitude (default 40.57)",
        type=str,
        required=False,
        default="40.57",
    )
    parser.add_argument(
        "-n",
        "--longitude",
        help="Longitude (default -105.23)",
        type=str,
        required=False,
        default="-105.23",
    )
    parser.add_argument(
        "-y",
        "--year",
        help="Year (default today)",
        type=int,
        required=False,
        default=YEAR,
    )
    parser.add_argument(
        "-m",
        "--month",
        help="Month (default today)",
        type=int,
        required=False,
        default=MONTH,
    )
    parser.add_argument(
        "-d",
        "--day",
        help="Day (default today)",
        type=int,
        required=False,
        default=DAY,
    )
    parser.add_argument(
        "-c",
        "--confidence",
        help="Minimum Confidence (default 0.25)",
        type=float,
        required=False,
        default="0.25",
    )

    args = parser.parse_args()

    handler_factory = functools.partial(
        TCPHandler,
        lat=args.latitude,
        lon=args.longitude,
        year=args.year,
        month=args.month,
        day=args.day,
        confidence=args.confidence,
    )

    initialize_logging()

    class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
        pass

    server = ThreadedTCPServer((args.ip, args.port), handler_factory)

    try:
        n = datetime.datetime.now(datetime.timezone.utc).astimezone()
        print(n, "sndid-server started on", args.ip + ":" + str(args.port))
        logging.info(
            n.strftime("%Y-%m-%d %H:%M:%S")
            + " sndid-server started on "
            + args.ip
            + ":"
            + str(args.port)
        )
        server.serve_forever()
    except KeyboardInterrupt:
        server.server_close()


def main():
    start_server()


if __name__ == "__main__":
    main()



# sndid/detectionswriter.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import queue
import threading
from .textwriter import TextWriter
from .audiowriter import AudioWriter


class DetectionsWriter:
    def __init__(self, detections_file, detections_dir):
        logging.debug("Initializing DetectionsWriter")
        self.text_writer = TextWriter(detections_file)
        self.audio_writer = AudioWriter(detections_dir)
        self.detection_queue = queue.Queue()
        self._stop_event = threading.Event()
        self._writer_thread = threading.Thread(target=self._write_detections_loop)
        self._writer_thread.daemon = True
        logging.debug("Starting DetectionsWriter thread.")
        self._writer_thread.start()

    def write_detections(self, detections, detection_timestamp, input_index):
        self.detection_queue.put((detections, detection_timestamp, input_index))
        logging.debug(f"DetectionsWriter detections added to queue: {len(detections)}")

    def _write_detections_loop(self):
        while not self._stop_event.is_set():
            try:
                detections, detection_timestamp, input_index = self.detection_queue.get(
                    timeout=1
                )
                logging.debug("Detections received from queue to DetectionsWriter.")
            except queue.Empty:
                continue
            self.text_writer.write_detections(
                detections, detection_timestamp, input_index
            )

    def write_audio(self, recording_data, detections, detection_timestamp, input_index):
        self.audio_writer.write_audio(
            recording_data, detections, detection_timestamp, input_index
        )
        logging.debug("Audio written from DetectionsWriter.")

    def stop(self):
        self._stop_event.set()
        logging.debug("Waiting for DetectionsWriter thread to finish.")
        self._writer_thread.join()
        logging.debug("DetectionsWriter thread stopped.")



# sndid/config.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import os
import yaml
from .constants import (
    N_INPUTS,
    SEGMENT_LENGTH,
    MIN_CONFIDENCE,
    LAT,
    LON,
    LOG_LEVEL,
    SAMPLE_RATE,
    OUTPUT_DIR,
)


class Config:
    @staticmethod
    def load_config(file_path=None):
        config = {}
        if file_path:
            with open(file_path, "r") as file:
                config = yaml.safe_load(file)
        return Config.set_defaults(config)

    @staticmethod
    def set_defaults(config):
        config["n_inputs"] = int(
            os.getenv("N_INPUTS", config.get("n_inputs", N_INPUTS))
        )
        config["segment_length"] = int(
            os.getenv("SEGMENT_LENGTH", config.get("segment_length", SEGMENT_LENGTH))
        )
        config["min_confidence"] = float(
            os.getenv("MIN_CONFIDENCE", config.get("min_confidence", MIN_CONFIDENCE))
        )
        config["lat"] = float(os.getenv("LAT", config.get("lat", LAT)))
        config["lon"] = float(os.getenv("LON", config.get("lon", LON)))
        config["log_level"] = os.getenv("LOG_LEVEL", config.get("log_level", LOG_LEVEL))
        config["output_dir"] = os.getenv(
            "OUTPUT_DIR", config.get("output_dir", OUTPUT_DIR)
        )
        config["sample_rate"] = int(
            os.getenv("SAMPLE_RATE", config.get("sample_rate", SAMPLE_RATE))
        )
        return config



# sndid/jack_lazy_loader.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import importlib
import logging

_import_cache = {}


def jack_lazy_load(module_name):
    logging.debug(f"Starting lazy load for: {module_name}")
    if module_name in _import_cache:
        logging.debug(f"Returning cached version of {module_name}")
        return _import_cache[module_name]
    logging.debug(f"Importing and caching {module_name} for the first time")
    module = importlib.import_module(module_name)
    _import_cache[module_name] = module
    logging.debug(f"{module_name} successfully imported and cached")
    return module



# sndid/segment_analyzer.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import threading
import logging


class SegmentAnalyzer(threading.Thread):
    def __init__(self, analyzer_wrapper):
        logging.debug("Initialize SegmentAnalyzer")
        self.analyzer_wrapper = analyzer_wrapper

    def analyze(self, input_index, segment):
        logging.debug(f"SegmentAnalyzer Analyzing segment for input {input_index}")
        recording, detection_timestamp = self.analyzer_wrapper.analyze(
            input_index, segment
        )
        if recording.detections:
            logging.debug(
                f"SegmentAnalyzer detections found in segment for input {input_index}: {recording.detections}"
            )
            return segment, recording.detections, detection_timestamp
        return None



# sndid/analyzerwrapper.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import io
from contextlib import redirect_stdout, redirect_stderr
from datetime import datetime
from .jack_lazy_loader import jack_lazy_load


class AnalyzerWrapper:
    def __init__(self, config):
        logging.debug("Initializing AnalyzerWrapper")
        Analyzer = jack_lazy_load("birdnetlib.analyzer").Analyzer
        self.config = config
        self.f = io.StringIO()
        self.analyzer = Analyzer()

    def analyze(self, input_index, segment):
        logging.debug(f"Running analyzer for input {input_index + 1}")
        RecordingBuffer = jack_lazy_load("birdnetlib").RecordingBuffer
        with redirect_stdout(self.f), redirect_stderr(self.f):
            recording_timestamp = datetime.now()
            logging.debug(
                f"Creating RecordingBuffer instance for input {input_index + 1}"
            )
            recording = RecordingBuffer(
                self.analyzer,
                segment,
                self.config["sample_rate"],
                lat=self.config["lat"],
                lon=self.config["lon"],
                date=recording_timestamp,
                min_conf=self.config["min_confidence"],
            )
            logging.debug(f"Analyzing RecordingBuffer for input {input_index + 1}")
            recording.analyze()
        if recording.detections:
            logging.info(
                f"Detections for input {input_index + 1}: {recording.detections}"
            )
        return recording, recording_timestamp



# sndid/constants.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0

# Default values.
# Can be overrided by configuration file or command line options.
import os

N_INPUTS = 2
SEGMENT_LENGTH = 240000
MIN_CONFIDENCE = 0.25
LAT = 40
LON = -105
LOG_LEVEL = "INFO"
SAMPLE_RATE = 48000
OUTPUT_DIR = os.path.expanduser("~/detections")



# sndid/audiowriter.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import logging
import os
from .jack_lazy_loader import jack_lazy_load


class AudioWriter:
    def __init__(self, detections_dir):
        self.detections_dir = detections_dir
        logging.debug("Initialized AudioWriter with detections_dir: %s", detections_dir)

    def write_audio(self, recording_data, detections, detection_timestamp, input_index):
        for d in detections:
            start_sample = int(d["start_time"] * 48000)
            end_sample = int(d["end_time"] * 48000)
            logging.debug(
                "Processing detection with start_sample: %s and end_sample: %s",
                start_sample,
                end_sample,
            )
            detection_audio = recording_data[start_sample:end_sample]
            species_dir = os.path.join(
                self.detections_dir,
                f"input_{input_index + 1}",
                d["scientific_name"].lower().replace(" ", "_"),
            )

            logging.debug("species_dir: %s", species_dir)
            os.makedirs(species_dir, exist_ok=True)
            detection_filename = f"{d['scientific_name'].lower().replace(' ', '_')}_{detection_timestamp.strftime('%Y-%m-%d_%H-%M-%S')}.wav"
            detection_filepath = os.path.join(species_dir, detection_filename)
            logging.debug("detection_filepath: %s", detection_filepath)
            wavfile = jack_lazy_load("scipy.io.wavfile")
            wavfile.write(detection_filepath, 48000, detection_audio)
            logging.debug("Audio file written: %s", detection_filename)



# sndid/sndidjack.py
# Copyright 2024, Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import concurrent.futures
import io
import jack
import json
import logging
import numpy as np
import os
import queue
import sys
import threading
from .audiodetectionwriter import AudioDetectionWriter
from .segment_analyzer import SegmentAnalyzer
from .detectionswriter import DetectionsWriter
from .analyzerwrapper import AnalyzerWrapper
from .__version__ import __version__


class SndIdJack:
    def __init__(self, config):
        logging.debug(f"{sys.argv[0].split('/')[-1]} {__version__}")
        self.config = config
        logging.debug(f"Configuration loaded: {json.dumps(self.config, indent=4)}")
        self.segment_length = self.config["segment_length"]
        self.n_inputs = self.config["n_inputs"]
        self.sample_rate = self.config["sample_rate"]
        self.detections_file = os.path.join(self.config["output_dir"], "detections.txt")
        self.detections_dir = self.config["output_dir"]
        logging.debug(f"Creating detection writer for file: {self.detections_file}")
        self.detection_writer = AudioDetectionWriter(
            self.detections_file, self.detections_dir
        )
        self.min_confidence = config["min_confidence"]
        self.lat = config["lat"]
        self.lon = config["lon"]
        self.f = io.StringIO()
        logging.debug(f"Initializing {self.n_inputs} segment buffers")
        self.segments = [np.ndarray([1024]) for _ in range(self.n_inputs)]
        self.data_queues = [queue.Queue(maxsize=10) for _ in range(self.n_inputs)]
        logging.debug("Setting up thread event")
        self.event = threading.Event()
        self.clientname = "SndIdJack"
        self.client = jack.Client(self.clientname)
        logging.debug(f"Creating ThreadPoolExecutor with max_workers: {8}")
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=8)
        self.detections_writer = DetectionsWriter(
            self.detections_file, self.detections_dir
        )
        self.analyzer_wrapper = AnalyzerWrapper(self.config)

    def register_client(self):
        logging.debug("Registering client inports")
        for i in range(1, self.n_inputs + 1):
            self.client.inports.register(f"input_{i}")

    def segment_data(self, input_index):
        while not self.data_queues[input_index].empty():
            data = self.data_queues[input_index].get()
            self.segments[input_index] = np.append(
                self.segments[input_index], data, axis=0
            )
            if len(self.segments[input_index]) >= self.segment_length:
                segment = self.segments[input_index][: self.segment_length]
                self.executor.submit(self.analyze_segment, input_index, segment)
                logging.debug(
                    f"Submit segment for analysis in a new thread for input {input_index + 1}."
                )
                self.segments[input_index] = self.segments[input_index][
                    self.segment_length :
                ]

    def analyze_segment(self, input_index, segment):
        result = self.segment_analyzer.analyze(input_index, segment)
        if result:
            segment, detections, detection_timestamp = result
            self.detection_writer.detection_queue.put(
                (segment, detections, detection_timestamp, input_index)
            )

    def connect_client(self):
        logging.debug("Connecting client")
        with self.client:
            for i in range(self.n_inputs):
                self.client.connect(
                    f"ardour:Master/audio_out {i+1}", f"SndIdJack:input_{i+1}"
                )
            logging.info("Press Ctrl+C to stop")
            try:
                self.event.wait()
            except KeyboardInterrupt:
                logging.info("Interrupted by user")

    def process(self, frames):
        if len(self.client.inports) != self.n_inputs:
            self._update_data_structures()
        for i, port in enumerate(self.client.inports):
            data = port.get_array()
            self.data_queues[i].put(data)
            self.segment_data(i)

    def _update_data_structures(self):
        self.n_inputs = len(self.client.inports)
        while len(self.segments) < self.n_inputs:
            self.segments.append(np.ndarray([1024]))
        while len(self.data_queues) < self.n_inputs:
            self.data_queues.append(queue.Queue(maxsize=10))

    def shutdown(self, status, reason):
        logging.debug("JACK shutdown")
        logging.debug(f"Status: {status}")
        logging.debug(f"Reason: {reason}")
        self.event.set()

    def main(self):
        @self.client.set_process_callback
        def process_callback(frames):
            # logging.debug("process callback")
            self.process(frames)

        @self.client.set_shutdown_callback
        def shutdown_callback(status, reason):
            logging.debug("shuddown callback")
            self.shutdown(status, reason)

        logging.debug("Setting up segment analyzer and registering client")
        self.segment_analyzer = SegmentAnalyzer(self.analyzer_wrapper)
        self.register_client()
        self.connect_client()



# sndid/stream.py
# Copyright 2023, 2024 Jeff Moe <moe@spacecruft.org>
# Licensed under the Apache License, Version 2.0
import argparse
import signal
import subprocess
import sys


def signal_handler(signum, frame):
    print("\nReceived signal to exit. Cleaning up...")
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


def run_ffmpeg_command(ip, port, time, url):
    """Runs the ffmpeg command with subprocess, handling piping correctly."""
    ffmpeg_cmd = [
        "ffmpeg",
        "-hide_banner",
        "-loglevel",
        "quiet",
        "-i",
        url,
        "-acodec",
        "pcm_s16le",
        "-ac",
        "1",
        "-vcodec",
        "vnull",
        "-f",
        "wav",
        "-t",
        time,
        "-",
    ]

    nc_cmd = ["nc", "-q", "0", ip, port]

    ffmpeg_process = subprocess.Popen(ffmpeg_cmd, stdout=subprocess.PIPE)
    subprocess.run(nc_cmd, stdin=ffmpeg_process.stdout, text=True)
    ffmpeg_process.wait()

    if ffmpeg_process.returncode != 0:
        raise subprocess.CalledProcessError(ffmpeg_process.returncode, ffmpeg_cmd)


def main():
    parser = argparse.ArgumentParser(description="Run sndid-stream")
    parser.add_argument(
        "-i",
        "--ip",
        help="Server IP address (default 127.0.0.1)",
        type=str,
        required=False,
        default="127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Server network port (default 9988)",
        type=int,
        required=False,
        default="9988",
    )
    parser.add_argument(
        "-t",
        "--time",
        help="Length of segments in seconds (default 60)",
        type=int,
        required=False,
        default="60",
    )
    parser.add_argument("-u", "--url", help="Input url", type=str, required=True)

    args = parser.parse_args()
    IP = args.ip
    PORT = str(args.port)
    TIME = str(args.time)
    URL = args.url

    try:
        while True:
            print("Sending stream...")
            run_ffmpeg_command(IP, PORT, TIME, URL)
    except KeyboardInterrupt:
        print("Exiting on user interrupt.")
    finally:
        print("Script exit cleanup.")


if __name__ == "__main__":
    main()
